/**
  *  \file client/chart.cc
  *  \brief Starchart Drawing
  *
  *  This module contains the routines to draw the starchart. It is a
  *  more-or-less straightforward port of the PCC 1.0.17 starchart
  *  code.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <climits>
#include "client/chart.h"
#include "client/marks.h"
#include "cpluslib/bitops.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/ship.h"
#include "gfx/clip.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "game/ufo.h"

#if 0
# include "gfx/gfxaa.h"
# define drawCircle drawCircleAA
# define drawLine   drawLineAA
#endif

static const int
    SCRingRadius     = 3,       ///< Radius of planet ring, in ly.
    SCMaxRingRadius  = 6,       ///< Max.\ radius of scaled planet ring, in pixels.
    SCCrossRadius    = 6,       ///< Size of starbase cross, in ly.
    SCMaxCrossRadius = 12,      ///< Max.\ radius of scaled SB cross, in pixels.
    SCObjCrossRadius = 5,       ///< Size of cross in center of a circular object (mine/UFO), in ly.
    SCMaxObjCrossRadius = 10;   ///< Max.\ size of scaled cross in center of a circular object.

static const uint8 ionstorm_init[] = { 0x88, 0x00, 0x22, 0x00,
                                       0x88, 0x00, 0x22, 0x00 };
static const uint8 ionstorm_dense_init[] = { 0x88, 0x44, 0x22, 0x44,
                                             0x88, 0x44, 0x22, 0x44 };
static const uint8 ufofill_init[] = { 0x88, 0x55, 0x22, 0x55,
                                      0x88, 0x55, 0x22, 0x55 };

const GfxFillPattern IONSTORM_FILL       = ionstorm_init;
const GfxFillPattern IONSTORM_DENSE_FILL = ionstorm_dense_init;
const GfxFillPattern UFO_FILL            = ufofill_init;

/** Draw a, possibly filled, circle.
    \param f    zero for outline circle, nonzero for filled. */
static void
drawObject(GfxCanvas& can, GfxParameters& parm, GfxPoint pt, int r, int32 f)
{
    if (f) {
        drawFilledCircle(can, parm, pt.x, pt.y, r);
    } else {
        drawCircle(can, parm, pt.x, pt.y, r);
    }
}

/** Set fill style for drawing a minefield. Must frob the pattern a bit
    to make the minefield appear the same in all positions of the map. */
static void
setMineFillStyle(GfxParameters& parm, GfxPoint pt, GUniverse& /*univ*/,
                 GMinefield& pmf)
{
    if (pmf.getOwner() == ::getPlayerId())
        /* avoid that own and foreign minefields 'overwrite' each other */
        ++pt.y;
    if (pmf.isWeb()) {
        parm.fill_pattern = GRAY50_FILL;
        if ((pt.x + pt.y) & 1)
            parm.fill_pattern.shiftUp(1);
    } else {
        parm.fill_pattern = GRAY25_FILL;
        parm.fill_pattern.shiftUp((15 + 2*(pt.x & 1) - pt.y & 3) & 3);
    }
}

/***************************** GChartViewport ****************************/

/*! \class GChartViewport
    This encapsulates the logic for drawing a starchart. Starcharts
    can be displayed in various zoom levels specified as fractions. */

/** Initialize. Call after setting the other parameters to compute
    derived information. */
void
GChartViewport::init()
{
    center_screen.x = view.x + view.w / 2;
    center_screen.y = view.y + view.h / 2;
    origin_screen.x = center_screen.x - scale(center.x);
    origin_screen.y = center_screen.y + scale(center.y);
}

/** Scale. Convert game distance into screen distance. The value n
    should be chosen so that the result is in range permissible for an
    int16 (because int may have 16 bits only, and drawing code might
    only be able to handle 16 bit coordinates). */
int
GChartViewport::scale(int n) const
{
    /* we want to round towards -INF instead of 0 */
    if ((-1)/2 == 0 && n < 0)
        return - (-long(n) * mult) / divi;
    return long(n) * mult / divi;
}

/** Scale, with Clipping. Compute scale(n)+add, but clip the result to
    be in the guaranteed range of int16. Unlike scale(), this is designed
    to handle out-of-range values of n. */
int
GChartViewport::scaleClip(int n, int add) const
{
    /* no need to fix rounding behaviour, this one's called to
       scale large values... */
    long l = long(n) * mult / divi + add;
    if (l < -0x7FFF)
        return -0x7FFF;
    else if (l > 0x7FFF)
        return 0x7FFF;
    else
        return l;
}

/** Unscale. Convert screen distance into game distance. */
int
GChartViewport::unscale(int n) const
{
    if ((-1)/2 == 0 && n < 0)
        return - (-long(n) * divi) / mult;
    return long(n) * divi / mult;
}

/** Draw Starchart, main entry point.
    \param can    Canvas to draw to
    \param opt    Options (what to draw)
    \param univ   Universe to display
    \param bbox   Applicable bounding box */
void
GChartViewport::draw(GfxCanvas& can, GChartOptions& opt, GUniverse& univ,
                     const GChartBBox& bbox)
{
    GfxClipRectFilter worker(view, can);

    init();
    // FIXME: set font....
    // FIXME: this order of starchart drawing has the drawback that in
    // wrapped mode, decoration *can* obscure ships (the "east" image
    // is drawn after the "center" image, hence an "east" planet ring
    // can obscure a "center" ship when they're close enough to the
    // border).
    // FIXME: exception safety regarding origin
    if (chart_conf.wrapped) {
        const GfxPoint origin_save = origin_screen;
        for (int i = -1; i <= 1; ++i) {
            for (int j = -1; j <= 1; ++j) {
                origin_screen.x = origin_save.x + scale(i*chart_conf.xsize);
                origin_screen.y = origin_save.y + scale(j*chart_conf.ysize);
                drawAux(worker, opt, univ, bbox);
            }
        }
        origin_screen = origin_save;
    } else {
        drawAux(worker, opt, univ, bbox);
    }
}

/** Draw moving object.
    \param can     canvas
    \param parm    graphics parameters to use
    \param pt      center point, screen coordinates
    \param radius  radius, screen coordinates
    \param speed   warp factor
    \param heading heading angle, in degrees
    \param f       nonzero to fill the object (options.fill) */
void
GChartViewport::drawMovingObject(GfxCanvas& can, GfxParameters& parm,
                                 GfxPoint pt, int radius,
                                 int speed, int heading, int32 f)
{
    drawObject(can, parm, pt, radius, f);
    if (speed > 0 && speed <= 20 && heading >= 0) {
        double h = heading * PI / 180;
        pt.x += roundToInt(radius * std::sin(h));
        pt.y -= roundToInt(radius * std::cos(h));

        int way = scale(speed * speed);
        int dx = roundToInt(way * std::sin(h));
        int dy = roundToInt(way * std::cos(h));

        drawArrow(can, parm, pt.x, pt.y, pt.x + dx, pt.y - dy,
                  2*mult < divi ? 3 : 5);
    }
}

/** Draw starchart, helper routine. This does the actual drawing of
    one starchart image. */
void
GChartViewport::drawAux(GfxCanvas& can, GChartOptions& opt, GUniverse& univ,
                        const GChartBBox& bbox)
{
    int range_x1 = scaleClip(bbox.getMinimumXY().x, origin_screen.x);
    int range_x2 = scaleClip(bbox.getMaximumXY().x, origin_screen.x);
    int range_y1 = scaleClip(-bbox.getMinimumXY().y, origin_screen.y);
    int range_y2 = scaleClip(-bbox.getMaximumXY().y, origin_screen.y);
    if (!can.isVisible(GfxRect(range_x1,             range_y2,
                               range_x2 - range_x1,  range_y1 - range_y2)))
        return;

    drawSectors(can, opt.show);
    if (opt.show & GChartOptions::co_Mine)
        drawMines(can, univ, opt.fill);
    if (opt.show & GChartOptions::co_Ufo)
        drawUfos(can, univ, opt.fill);
    if (opt.show & GChartOptions::co_Ion)
        drawIons(can, univ, opt.fill);
    if (opt.show & GChartOptions::co_Drawings)
        drawDrawings(can, univ, opt.show);
    drawShipSelAndVectors(can, univ, opt.show);
    drawPlanets(can, univ, opt.show);
    drawShips(can, univ, opt.show);
}

/** Draw a cross, used for the center of Ufos and mine fields. */
void
GChartViewport::drawCross(GfxCanvas& can, GfxParameters& parm, GfxPoint pt)
{
    int n = scale(SCObjCrossRadius);
    if (n <= 0)
        n = 1;
    else if (n > SCMaxObjCrossRadius)
        n = SCMaxObjCrossRadius;
    drawHLine(can, parm, pt.x - n, pt.y, pt.x + n);
    drawVLine(can, parm, pt.x, pt.y - n, pt.y + n);
}

/** Draw ion storms. Call only if ion storm drawing is actually enabled.
    \param can canvas to use
    \param univ universe
    \param fill options.fill */
void
GChartViewport::drawIons(GfxCanvas& can, GUniverse& univ, int fill)
{
    GfxParameters parm;
    parm.fill_pattern = IONSTORM_FILL;
    for (int sid = 1; sid <= NUM_STORMS; ++sid) {
        GIonStorm& i = univ.getIonStorm(sid);
        if (!i.isExistant())
            continue;

        if (i.getVoltage() < 50)
            parm.color = standard_colors[COLOR_BLUE];
        else if (i.getVoltage() < 100)
            parm.color = standard_colors[COLOR_GRAY];
        else if (i.getVoltage() < 150)
            parm.color = standard_colors[COLOR_WHITE];
        else
            parm.color = standard_colors[COLOR_RED];

        drawMovingObject(can, parm,
                         scaleXY(i.getPos()),
                         scale(i.getRadius()),
                         i.getSpeed(), i.getHeading(),
                         fill & GChartOptions::co_Ion);
    }
}

/** Draw user drawings. Call only if these are actually enabled.
    \param can  Canvas to draw on
    \param trn  Turn to get drawings from
    \param show options.show, to test whether labels should be drawn */
void
GChartViewport::drawDrawings(GfxCanvas& can, GUniverse& trn, int show)
{
    GfxParameters parm;
    GDrawingContainer& d = trn.getDrawings();

    parm.just_x = 1;
    parm.just_y = 0;
    parm.font = fonts[getFont()];

    for (GDrawingContainer::iterator i = d.begin(); i != d.end(); ++i) {
        if (i->isVisible()) {
            // FIXME: color remapping and flashing
            parm.color = standard_colors[i->getColor()];

            GfxPoint origin = scaleXY(i->getPos());
            switch (i->getType()) {
             case LineDrawing: {
                GfxPoint end = scaleXY(i->getPos2());
                drawLine(can, parm, origin.x, origin.y, end.x, end.y);
                break;
             }
             case RectangleDrawing: {
                GfxPoint end = scaleXY(i->getPos2());
                drawRectangle(can, parm, GfxRect(std::min(origin.x, end.x),
                                                 std::min(origin.y, end.y),
                                                 std::abs(origin.x - end.x),
                                                 std::abs(origin.y - end.y)));
                break;
             }
             case CircleDrawing:
                drawCircle(can, parm, origin.x, origin.y, scale(i->getCircleRadius()));
                break;
             case MarkerDrawing:
                if (divi < 2*mult) {
                    /* draw marker */
                    const TMarkerImage img = getUserMarker(i->getMarkerKind(), true);
                    drawMarker(can, parm, img, origin.x, origin.y);

                    /* draw text */
                    if (show & GChartOptions::co_Labels) {
                        string_t text = strFirst(i->getComment(), "|");
                        if (text.size())
                            outTextF(can, parm, origin.x, origin.y + img.height, 600, text);
                    }
                } else {
                    drawMarker(can, parm, getUserMarker(i->getMarkerKind(), false), origin.x, origin.y);
                }
            }
        }
    }

    GExplosionContainer& e = trn.getExplosions();
    for (GExplosionContainer::iterator i = e.begin(); i != e.end(); ++i) {
        GfxPoint pos = scaleXY(i->getPos());
        bool big = divi < 2*mult;
        /* red '+' */
        parm.color = standard_colors[COLOR_RED];
        drawMarker(can, parm, getUserMarker(0, big), pos.x, pos.y);
        /* yellow 'x' */
        parm.color = standard_colors[COLOR_YELLOW];
        drawMarker(can, parm, getUserMarker(2, big), pos.x, pos.y);
    }
}

/** Draw Minefields. Call only if minefield drawing is actually enabled.
    \param can   canvas to draw to
    \param univ   turn to display
    \param fill  options.fill */
void
GChartViewport::drawMines(GfxCanvas& can, GUniverse& univ, int fill)
{
    GfxParameters parm;
    for (GUniverse::MinefieldMap::iterator i = univ.getMinefieldBegin();
         i != univ.getMinefieldEnd(); ++i)
    {
        GMinefield& pmf = *i->second;
        GfxPoint pt = scaleXY(pmf.getPos());
        setMineFillStyle(parm, origin_screen, univ, pmf);

        GPlayerRelation rel = getPlayerRelation(pmf.getOwner());
        if (rel == is_Me)
            parm.color = standard_colors[COLOR_GREENSCALE+4];
        else if (rel == is_Ally)
            parm.color = standard_colors[COLOR_DARKYELLOWSCALE+4];
        else
            parm.color = standard_colors[COLOR_DARK];
        
        drawObject(can, parm, pt, scale(pmf.getRadius()), fill & GChartOptions::co_Mine);
        drawCross(can, parm, pt);
    }
}

void
GChartViewport::drawUfos(GfxCanvas& can, GUniverse& trn, int fill)
{
    GfxParameters parm;
    const GUfoContainer& uc = trn.getUfos();
    for (GUfoContainer::const_iterator i = uc.begin(); i != uc.end(); ++i) {
        const GUfo& ufo = *i;
        if (ufo.isExistant()) {
            /* set draw style */
            GfxPoint pt = scaleXY(ufo.getPos());
            parm.fill_pattern = UFO_FILL;
            parm.fill_pattern.shiftDown(pt.y & 3);
            parm.fill_pattern.shiftRight(pt.x & 3);
            parm.color = standard_colors[ufo.getColor()];

            /* draw it */
            drawMovingObject(can, parm, pt,
                             scale(ufo.getRadius()),
                             ufo.getSpeed(),
                             ufo.getHeading(),
                             fill & GChartOptions::co_Ufo);
            drawCross(can, parm, pt);
        }
    }
}

/** Draw sector/starchart borders.
    \param can    output canvas
    \param show   options.show */
void
GChartViewport::drawSectors(GfxCanvas& can, int show)
{
    GfxParameters parm;
    int  dx = std::min(int(chart_conf.xsize / 200), 10);
    int  dy = std::min(int(chart_conf.ysize / 200), 10);
    int  center_x = (chart_conf.xmin + chart_conf.xmax) / 2;
    int  center_y = (chart_conf.ymin + chart_conf.ymax) / 2;

    parm.color = standard_colors[COLOR_DARK];
    if (show & GChartOptions::co_Sectors) {
        parm.line_style = (origin_screen.y & 1) ? 0x55 : 0xAA;
        for (int i = -dx; i <= dx; ++i)
            drawVLine(can, parm,
                      scaleX(center_x + 100*i),
                      scaleY(center_y - 100*dy),
                      scaleY(center_y + 100*dy));

        parm.line_style = (origin_screen.x & 1) ? 0x55 : 0xAA;
        for (int i = -dy; i <= dy; ++i)
            drawHLine(can, parm,
                      scaleX(center_x - 100*dy),
                      scaleY(center_y + 100*i),
                      scaleX(center_x + 100*dy));
    }

    if (show & GChartOptions::co_Borders) {
        parm.line_style = ror8(0x27, origin_screen.y);
        drawVLine(can, parm,
                  scaleX(chart_conf.xmin),
                  scaleY(chart_conf.ymin),
                  scaleY(chart_conf.ymax));
        drawVLine(can, parm,
                  scaleX(chart_conf.xmax),
                  scaleY(chart_conf.ymin),
                  scaleY(chart_conf.ymax));

        parm.line_style = ror8(0x27, origin_screen.x);
        drawHLine(can, parm,
                  scaleX(chart_conf.xmin),
                  scaleY(chart_conf.ymin),
                  scaleX(chart_conf.xmax));
        drawHLine(can, parm,
                  scaleX(chart_conf.xmin),
                  scaleY(chart_conf.ymax),
                  scaleX(chart_conf.xmax));
    }
}

/** Draw Ships, part I. Draws ship selection and trails. These are
    drawn below the planets, to not obscure other stuff. */
void
GChartViewport::drawShipSelAndVectors(GfxCanvas& can, GUniverse& univ, int show)
{
    if (show & GChartOptions::co_Selection) {
        GfxParameters parm;
        parm.color = standard_colors[COLOR_YELLOW];
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            GShip& s = univ.getShip(i);
            if (s.isExistant() && s.isMarked()) {
                GfxPoint pt = scaleXY(s.getPos());
                drawSelection(can, parm, pt.x, pt.y, mult, divi);
            }
        }
    }
    if (!(show & GChartOptions::co_ShipDots)) {
        GfxParameters parm;
        const bool big = (divi < 2*mult);
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            const GShip& s = univ.getShip(i);
            if (s.isExistant() && s.getOrbit() == 0) {
                GPlayerRelation rel = getPlayerRelation(s.getOwner());
                parm.color = standard_colors[rel == is_Me ? COLOR_GREEN :
                                             rel == is_Ally ? COLOR_YELLOW : COLOR_RED];
                GfxPoint pt = scaleXY(s.getPos());
                drawShipIcon(can, parm, rel, big, pt.x, pt.y);
            }
        }
    }
//   IF show AND co_Trails<>0 THEN BEGIN
//     FOR i:=1 TO NUM_SHIPS DO
//      IF (Ships^[i].HHist<>NIL) THEN
//       NShowVector(view, i);
//   END ELSE IF view.vec_id>0 THEN
//    NShowVector(view, view.vec_id);
}

/** Draw Ships, part II. These are drawn above the planets, so that planet
    rings do not obscure ships in warp wells, like in VPA, for example
    (hint hint! ;-) */
void
GChartViewport::drawShips(GfxCanvas& can, GUniverse& univ, int show)
{
    (void) show;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = univ.getShip(i);
        if (! s.isExistant())
            continue;

        GfxPoint pt = scaleXY(s.getPos());
        int color = BLACK_COLOR_SET[getPlayerColor(s.getOwner())];
        if ((show & GChartOptions::co_ShipDots) != 0 && s.getOrbit() == 0)
            can.drawPixel(pt.x, pt.y, standard_colors[color], GFX_ALPHA_OPAQUE);
// FIXME:
//     IF (Fleets^[i].FleetID = i) AND (2*view.mult > view.divi) THEN BEGIN
//       PutPixel(x-1, y-1, Color);
//       PutPixel(x-1, y+1, Color);
//       PutPixel(x+1, y-1, Color);
//       PutPixel(x+1, y+1, Color);
//     END;
//     IF (Ships^[i].Text<>NIL) AND (show AND co_Labels<>0) THEN BEGIN
//       IF Ships^[i].Orbit<>0 THEN Inc(y, 4);
//       OutLTextXY(x, y+1, 600, Ships^[i].Text^);
//     END;
    }
}

/** Draw Warp Well.
    \param can Canvas to draw on
    \param pt Screen coordinates */
void
GChartViewport::drawWarpWell(GfxCanvas& can, GfxPoint pt)
{
    const int range = config.GravityWellRange();
    const int half = scale(1) / 2;

    GfxParameters parm;
    parm.color = standard_colors[COLOR_SHIELD+4];

    if (config.AllowGravityWells() && range > 0) {
        if (config.RoundGravityWells()) {
            /* This draws an edgy frame around the planet's warp well.
               At large zoom levels, each position in space becomes a
               "square" with an actual area; we draw the outline for
               these squares. For simplicity, the algorithm is split
               into 8 octants. */
            int wwx = range;
            int wwy = 0;
    
            int sx = scale(wwx) + half;
            int sy = scale(wwy) + half;
            drawVLine(can, parm, pt.x - sx, pt.y - sy, pt.y + sy);
            drawVLine(can, parm, pt.x + sx, pt.y - sy, pt.y + sy);
            drawHLine(can, parm, pt.x - sy, pt.y - sx, pt.x + sy);
            drawHLine(can, parm, pt.x - sy, pt.y + sx, pt.x + sy);

            while (wwx > wwy) {
                ++wwy;
                int nsy = scale(wwy) + half;
                if (square(wwx) + square(wwy) > square(range)) {
                    --wwx;
                    int nsx = scale(wwx) + half;
                    drawHLine(can, parm, pt.x - sx, pt.y - sy, pt.x - nsx);
                    drawHLine(can, parm, pt.x - sx, pt.y + sy, pt.x - nsx);
                    drawHLine(can, parm, pt.x + sx, pt.y - sy, pt.x + nsx);
                    drawHLine(can, parm, pt.x + sx, pt.y + sy, pt.x + nsx);
                    drawVLine(can, parm, pt.x - sy, pt.y + sx, pt.y + nsx);
                    drawVLine(can, parm, pt.x - sy, pt.y - sx, pt.y - nsx);
                    drawVLine(can, parm, pt.x + sy, pt.y + sx, pt.y + nsx);
                    drawVLine(can, parm, pt.x + sy, pt.y - sx, pt.y - nsx);
                    sx = nsx;
                }
                drawVLine(can, parm, pt.x - sx, pt.y - sy, pt.y - nsy);
                drawVLine(can, parm, pt.x - sx, pt.y + sy, pt.y + nsy);
                drawVLine(can, parm, pt.x + sx, pt.y - sy, pt.y - nsy);
                drawVLine(can, parm, pt.x + sx, pt.y + sy, pt.y + nsy);
                drawHLine(can, parm, pt.x - sy, pt.y - sx, pt.x - nsy);
                drawHLine(can, parm, pt.x - sy, pt.y + sx, pt.x - nsy);
                drawHLine(can, parm, pt.x + sy, pt.y - sx, pt.x + nsy);
                drawHLine(can, parm, pt.x + sy, pt.y + sx, pt.x + nsy);
                sy = nsy;
            }
        } else {
            /* Just a rectangle */
            const int dx = scale(range) + half;
            drawRectangle(can, parm, GfxRect(pt.x - dx, pt.y - dx, 2*dx+1, 2*dx+1));
        }
    }
}

/** Draw Planets. */
void
GChartViewport::drawPlanets(GfxCanvas& can, GUniverse& univ, int show)
{
    enum RingType { r_Enemy, r_Own, r_Unowned, r_Unknown };
    GfxParameters parm;
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GPlanet& p = univ.getPlanet(i);
        if (!p.isExistant())
            continue;

        GfxPoint pt = scaleXY(p.getPos());

        /* Selection */
        if ((show & GChartOptions::co_Selection) && (p.isMarked())) {
            parm.color = standard_colors[COLOR_YELLOW];
            drawSelection(can, parm, pt.x, pt.y, mult, divi);
        }

        bool     base;          // true iff planet has base
        RingType ring;          // type of planet ring
        int      incy;          // Y increment for text
        bool     shipg;         // true iff ship ring is dotted (guessed)
        int      shipring;      // color of ship ring, 0 if none

        /* Compute decoration -- Planet */
        // FIXME: history & stuff
        incy = 1;
        if (p.getOwner() == getPlayerId()) {
            ring = r_Own;
            base = p.hasBase();
        } else {
            if (p.getOwner())
                ring = r_Enemy;
            else if (p.hasAnyPlanetData())
                ring = r_Unowned;
            else
                ring = r_Unknown;
            base = false;
        }
        /* Compute decoration -- Ship */
        shipg = false;
        shipring = 0;
        switch(p.getStatus() & (GPlanet::FlagOwnShipsInOrbit + GPlanet::FlagEnemyShipsInOrbit + GPlanet::FlagAlliedShipsInOrbit)) {
         case 0:
            shipg = true;
            switch(p.getStatus() & (GPlanet::FlagGuessedEnemyInOrbit + GPlanet::FlagGuessedAllyInOrbit)) {
             case 0:
                shipring = 0;
                break;
             case GPlanet::FlagGuessedAllyInOrbit:
                shipring = BLACK_COLOR_SET[UIColor::tc_Yellow];
                break;
             case GPlanet::FlagGuessedEnemyInOrbit:
                shipring = COLOR_RED;
                break;
             default:
                shipring = COLOR_YELLOW;
                break;
            }
            break;
         case GPlanet::FlagOwnShipsInOrbit:
            shipring = COLOR_GREEN;
            break;
         case GPlanet::FlagEnemyShipsInOrbit:
            shipring = COLOR_RED;
            break;
         case GPlanet::FlagAlliedShipsInOrbit:
         case GPlanet::FlagAlliedShipsInOrbit + GPlanet::FlagOwnShipsInOrbit:
            shipring = BLACK_COLOR_SET[UIColor::tc_Yellow];
            break;
         default:
            shipring = COLOR_YELLOW;
            break;
        }

        /* Warp Wells */
        if (show & GChartOptions::co_WarpWells)
            drawWarpWell(can, pt);

        /* Draw the planet */
        if (ring != r_Own)
            parm.color = standard_colors[COLOR_BLUEBLACK];
        else
            parm.color = standard_colors[COLOR_BLUE];
        if (base) {
            int sbsize = std::min(int(scale(SCCrossRadius)), int(SCMaxCrossRadius));
            drawHLine(can, parm, pt.x - sbsize, pt.y, pt.x + sbsize);
            drawVLine(can, parm, pt.x, pt.y - sbsize, pt.y + sbsize);
        }
        can.drawPixel(pt.x, pt.y, standard_colors[COLOR_WHITE], GFX_ALPHA_OPAQUE);

        int r = std::min(int(scale(SCRingRadius)), int(SCMaxRingRadius));
        if (shipring == 0 && r == 0)
            r = 1;
        if (r) {
            if (ring == r_Enemy || ring == r_Own) {
                drawCircle(can, parm, pt.x, pt.y, r);
                incy += r;
            } else if (ring == r_Unowned) {
                drawDottedCircle(can, parm, pt.x, pt.y, r);
                incy += r;
            }
        }
        if (shipring != 0) {
            parm.color = standard_colors[shipring];
            ++r;
            if (shipg)
                drawDottedCircle(can, parm, pt.x, pt.y, r);
            else
                drawCircle(can, parm, pt.x, pt.y, r);
            incy = r+1;
        }
//     IF (Stars^[i].Text<>NIL) AND (show AND co_Labels<>0) THEN BEGIN
//       Color:=Gray;
//       OutLTextXY(x,y+IncY,600,Stars^[i].Text^);
//     END;
    }
}

int
GChartViewport::getFont() const
{
    // IF view.mult > view.divi THEN BEGIN
    //   IF view.mult > 2*view.divi THEN BEGIN
    //     TextStyle := 0;
    //     AAColor := 0;
    //   END ELSE TextStyle := 3;
    // END ELSE BEGIN
    //   IF 2*view.mult < view.divi THEN BEGIN
    //     IF 3*view.mult < view.divi THEN opt.show := opt.show AND NOT co_Labels;
    //     TextStyle := 4
    //   END ELSE BEGIN
    //     TextStyle := 2;
    //   END;
    // END;

    if (mult > divi) {
        if (mult > 2*divi) {
            /* more then 2:1: 22 pt */
            return FONT_TITLE;
        } else {
            /* 1:1 up to 2:1: 16 pt */
            return FONT_NORMAL;
        }
    } else {
        if (2*mult < divi) {
            /* smaller than 1:2 */
            // FIXME: should be 6 or 8 pt font, as in PCC 1.x */
            return FONT_SMALL;
        } else {
            /* 1:2 up to 1:1: 12 pt */
            return FONT_SMALL;
        }
    }
}

/** Zoom in.
    \param bbox applicable bounding box (limits available zoom factors)
    \returns true iff zoom factor changed */
bool
GChartViewport::zoomIn(GChartBBox& bbox)
{
    if (divi > 1)
        --divi;
    else if (mult < bbox.getMaxZoom() * divi)
        ++mult;
    else
        return false;           // can't zoom in further

    /* Zoom range may be exceeded. For example, old zoom 5:2, limit 4;
       new zoom 5:1 */
    if (mult > divi * bbox.getMaxZoom()) {
        mult = bbox.getMaxZoom();
        divi = 1;
    }
    return true;
}

/** Zoom out.
    \param bbox applicable bounding box
    \returns true iff zoom factor changed */
bool
GChartViewport::zoomOut(GChartBBox& bbox)
{
    // FIXME: incomplete
    (void) bbox;
    if (mult > 1) {
        --mult;
        return true;
    } else if (divi < GChartBBox::MAX_ZOOM) {
        ++divi;
        return true;
    } else {
        return false;
    }
}

static GChartOptions chart_options[3] = {
    { GChartOptions::co_Default, GChartOptions::co_Default },
    { GChartOptions::co_Default, GChartOptions::co_Default },
    { GChartOptions::co_Default, GChartOptions::co_Default }
};

/** Get starchart options for a particular chart widget.
    \param is_scanner  true if scanner in a control screen, false if main chart
    \param mult,divi   zoom settings

    Currently implements the PCC 1.x policy. */
GChartOptions&
getChartOpts(bool is_scanner, int mult, int divi)
{
    if (is_scanner)
        return chart_options[2];
    else if (mult < divi)
        return chart_options[0];
    else
        return chart_options[1];
}

/** Toggle an option (or set of options).
    \param which_ones options to toggle, set of co_XXX */
void
GChartOptions::toggleOptions(int32 which_ones)
{
    // First, tristate options. We have three states:
    //    show   fill
    //      0      ?
    //      1      0
    //      1      1
    int32 tristate_opts   = which_ones & co_Tristate;
    int32 shown_tristate  = tristate_opts & show;        /* shown filled or shown empty */
    int32 hidden_tristate = tristate_opts & ~show;       /* hidden */
    int32 empty_tristate  = shown_tristate & ~fill;      /* shown filled */
    int32 filled_tristate = shown_tristate & fill;       /* shown empty */

    show |= hidden_tristate;                             /* show hidden */
    fill &= ~hidden_tristate;

    fill |= empty_tristate;                              /* fill empty */

    show &= ~filled_tristate;                            /* hide filled */
    fill &= ~filled_tristate;

    // Then, normal options
    int32 boolean_opts = which_ones & ~co_Tristate;
    show ^= boolean_opts;
}
