/**
  *  \file client/wtile.cc
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wtile.h"
#include "ui/ui.h"
#include "ui/uiglobal.h"

/********************************* WTile *********************************/

/** \class WTile
    \brief Widget displaying certain data (hooray.)

    Tiles can display the following stuff:
    - a constant heading
    - some fixed data
    - some variable data

    This class provides an interface intended to make display of such
    stuff easier. It automatically manages update events. In addition,
    a tile can request to have some associated controls (buttons).

    To avoid excess redraws, data is partitioned. Each tile knows on
    which events it wants to be updated (\c what parameter to
    constructor). Update events (see update(int)) include this
    parameter; the tile is redrawn if any of the bits in our \c what
    is set in the event, or if \c what is zero (meaning that we always
    want to be updated) or the event is zero (meaning that we don't
    know what was updated and thus update everything just-in-case).

    \todo I'm not 100% sure whether the redraw optimisations are still
    needed or whether we can simply update everything.

    (In PCC 1.x, the distinguishing property of a "Tile" was that it
    was freely moveable. In PCC 2, all widgets are moveable.) */

/** Standard constructor.
    \param size  size of this tile
    \param id    widget Id of this tile
    \param when  when this tile wants to be updated (flags). Zero means
    update on all events. Otherwise, the tile will only be redrawn when
    any of the given flags is set in the update event. See class
    description. */
WTile::WTile(GfxPoint size, int id, int when)
    : UIWidget(id), size(size), update_when(when), state(sContentOnly)
{
    computeCoordinates();
}

/** Standard constructor. This sets the widget Id to zero.
    \param wi,he size of this tile
    \param when see other constructor */
WTile::WTile(int wi, int he, int when)
    : UIWidget(0), size(wi, he), update_when(when), state(sContentOnly)
{
    computeCoordinates();
}

/** Draw the tile. Do not override in WTile descendants. */
void
WTile::draw(GfxCanvas& can)
{
    if (state != sContentOnly) {
        getSkin().startDraw();
        setColor(UIColor::tc_Heading);
        getSkin().setSolid(false);
        getSkin().outTextF(can, FONT_TITLE, extent.x, extent.y, extent.w, heading);
    }
    if (state != sHeadingOnly) {
        getSkin().startDraw();
        drawFixed(can);
        getSkin().startDraw();
        drawData(can);
    }
    can.update(getExtent());
}

/** Update event handler. This function is hooked into the
    WControlScreen::sig_changed signal.
    \param n reason for update. Zero means not known (so update
    just in case), otherwise a bitfield of flags. */
void
WTile::update(int n)
{
    if (!n || !update_when || (n & update_when))
        update();
}

/** Update. This function immediately redraws the /variable data/
    portion of the tile. */
void
WTile::update()
{
    /* do nothing if tile is not displayed */
    if (!getOwner() || !hasState(st_Exposed))
        return;

    /* draw it */
    ObjectHolder h;
    GfxCanvas& can = getOwner()->getCanvasFor(this, h);
    UISkin& s = getSkin();

    s.startDraw();
    s.setSolid();
    can.lock();
    drawData(can);
    can.unlock();
    can.update(extent);
}

/** Draw fixed data. Override this in derived classes to display data
    which does not change during the lifetime of this tile. */
void
WTile::drawFixed(GfxCanvas& /*can*/)
{ }

/** Draw variable data. Override this in derived classes to display
    data which does change during the lifetime of the tile. */
void
WTile::drawData(GfxCanvas& /*can*/)
{ }

/** Create controls belonging to this tile. Override this if you want
    to associate some controls with the tile.

    Controls should be added to the group \c g. The group will not
    have a layout manager, so use the two-argument add() version. You
    can use the provided ObjectHolder to store objects. For
    convenience, you can use the WTile::addControl() function which
    automatically adjusts the coordinates. Do not assume anything
    about the lifetime of the controls.

    \default does nothing */
void
WTile::addControls(UIGroup& g, ObjectHolder& holder)
{
    (void) g;
    (void) holder;
}

/** Add control to group. You can use this function to add a control
    to a group in response to an addControls() request.
    \param g the group passed to addControls()
    \param r relative position
    \param w the widget you want to add
    \param rel_pos if true, the coordinates are relative to your
    data area. If false (default), coordinates are relative to
    the whole tile. */
void
WTile::addControl(UIGroup& g, GfxRect r, UIWidget& w, bool rel_pos)
{
    r.moveBy(extent.x, extent.y);
    if (rel_pos)
        r.moveBy(0, y_origin);
    g.add(r, w);
}

/** Draw background. Derived drawData()/drawFixed() can use this to clear
    out regions of the background. \c r is in relative coordinates. */
void
WTile::drawBackground(GfxCanvas& can, GfxRect r)
{
    r.moveBy(extent.x, extent.y + y_origin);
    r.intersect(extent);
    getSkin().drawBackground(can, r);
}

/** Get Layout Information. Do not override in WTile descendants. */
void
WTile::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint size = this->size;
    if (state == sHeadingOnly)
        size.y = 0;
    size.y += y_origin;

    info.flags = 0;
    info.min_size = info.pref_size = size;
}

bool
WTile::handleEvent(const UIEvent& e, bool second_pass)
{
    return UIWidget::handleEvent(e, second_pass);
}

/** Get size of data area, with margins. This returns the dimensions
    of the data area you are expected to fill, in absolute coordinates.
    \param lmargin left margin (number of pixels to remove at the left)
    \param rmargin right margin */
GfxRect
WTile::getInnerSize(int lmargin, int rmargin)
{
    GfxRect rv = getExtent();
    rv.x += lmargin;
    rv.w -= lmargin+rmargin;
    rv.y += y_origin;
    rv.h -= y_origin;
    return rv;
}

/** Set display state.
    \todo Currently, this only works reliably when the tile is not
    yet in a group. It does not actually resize the tile. */
void
WTile::setDisplayState(State state)
{
    if (state == this->state)
        return;
    this->state = state;
    computeCoordinates();
    drawWidget();
    // FIXME: if widget is *really* already displayed, we probably
    // should tell our parent to resize us
}

/** Set heading of the tile.
    \return *this, for convenience */
WTile&
WTile::setHeading(string_t heading)
{
    this->heading = heading;
    drawWidget();
    return *this;
}

/** Compute coordinates, after a change to layout parameters. \internal */
void
WTile::computeCoordinates()
{
    if (state == sContentOnly) {
        y_origin = 0;
    } else {
        y_origin = font_heights[FONT_TITLE];
    }
}
