/**
  *  \file game/db.cc
  *  \brief Starchart Database I/O
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/db.h"
#include "game/drawing.h"
#include "game/global.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/nls.h"
#include "util/console.h"
#include "game/mine.h"
#include "game/planet.h"
#include "cpluslib/misc.h"
#include "cpluslib/format.h"

/** Starchart database record numbers. */
enum {
    rPlanetHistory  = 1,
    rShipHistory    = 2,
    rShipTrack      = 3,
    rMinefield      = 4,
    rPainting       = 5,
    rAutoBuild      = 6,
    rShipProperty   = 7,
    rPlanetProperty = 8,
    // PCC 1.1.6+:
    rShipScore      = 9,
    rPlanetScore    = 10,
    rPaintingTags   = 11,
    // PCC 1.1.7+:
    rUfoHistory     = 12
};

/** Load starchart database.
    \param s   Stream to read from
    \param trn Turn to populate with information */
void
loadChartDatabase(Stream& s, GGameTurn& trn)
{
    using namespace std;

    int ignored_entries = 0;

    // Read header
    char header[16];
    s.seek(0);
    if (s.read(header, sizeof(header)) != sizeof(header))
        throw FileFormatException(s, _("File too short"));
    if (memcmp(header, "CCchart\032", 8) != 0)
        throw FileFormatException(s, _("File is not a valid PCC chart database"));

    int16 turn              = getInt16(header, 8);
    uint16 start_offset     = getUint16(header, 10);
    uint16 num_planet_props = getUint16(header, 12);
    uint16 num_ship_props   = getUint16(header, 14);

    // FIXME: properties are skipped for now

    s.seek(start_offset);

    // Main loop
    GDrawingAtomMap atom_translation;

    char chunk_hdr[6];
    while (s.read(chunk_hdr, sizeof(chunk_hdr)) == sizeof(chunk_hdr)) {
        const int16 type = getInt16(chunk_hdr, 0);
        int32 size = getInt32(chunk_hdr, 2);        /* not const to allow count-down loops */
        const int32 start_pos = s.getPos();
        const int32 end_pos = start_pos + size;

        switch (type) {
         case rPlanetHistory: {
            /* a single planet record. Can have variable size, as we
               store more fields than PCC 1.x, and possibly extend it
               in the future. However, we expect a minimum size. */
            if (size >= 93) {
                /* 93 = size of planet data, plus timestamps */
                char tmp[TDbPlanet::size];
                zeroFill(tmp);
                if (size > TDbPlanet::size)
                    size = TDbPlanet::size;
                if (s.read(tmp, size) != size)
                    throw FileTooShortException(s);

                TDbPlanet dbp;
                dbp.known_to_have_natives = 0;
                getPartialStructure(tmp, dbp, size);
                if (dbp.planet.planet_id > 0 && dbp.planet.planet_id <= NUM_PLANETS)
                    trn.getCurrentUniverse().getPlanet(dbp.planet.planet_id).addHistoryData(dbp);
            } else {
                /* no known program writes these files */
            }
            break;
         }
         // case rShipHistory:
         // case rShipTrack:
         //    break;
         case rMinefield:
            while (size >= TDbMinefield::size) {
                /* read it */
                TDbMinefield dbm;
                if (!getStructure(s, dbm))
                    throw FileTooShortException(s);
                size -= TDbMinefield::size;

                /* enter into database */
                GMinefield* m = trn.getCurrentUniverse().getMinefield(dbm.id, GMinefield::mfa_None);
                if (dbm.turn > m->getTurnLastSeen())
                    m->setData(dbm.x, dbm.y, dbm.owner, dbm.units, dbm.turn, dbm.type & 1);
            }
            break;
         case rPainting: {
            SubStream ss(s);
            ss.setRange(start_pos, size);
            trn.getCurrentUniverse().getDrawings().load(ss, atom_translation, 0);
            atom_translation.clear();
            break;
         }
         case rAutoBuild: {
            int id = 0;
            while (size >= TAutobuildSettings::size && id < NUM_PLANETS) {
                /* read it */
                TAutobuildSettings abs;
                if (!getStructure(s, abs))
                    throw FileTooShortException(s);
                size -= TAutobuildSettings::size;

                /* enter into database */
                ++id;
                trn.getPlanet(id).setAutobuild(abs);
            }
            if (id != NUM_PLANETS || size != 0)
                console.write(LOG_WARN, _("Autobuild record has unexpected size"));
            break;
         }
         // case rShipProperty:
         // case rPlanetProperty:
         // case rShipScore:
         // case rPlanetScore:
         //    break;
         case rPaintingTags: {
            if (!atom_translation.isEmpty())
                console.write(LOG_WARN, _("Text record appears at unexpected place"));
            SubStream ss(s);
            ss.setRange(start_pos, size);
            atom_translation.clear();
            atom_translation.load(ss);
            break;
         }
         // case rUfoHistory:
         //    break;
         default:
            ++ignored_entries;
        }

        s.seek(end_pos);
    }
    if (ignored_entries != 0)
        console.write(LOG_INFO, format(_("%d database record%!1{s have%| has%} been ignored"), ignored_entries));
}

/*! \name Saving Starchart Database */
//@{
/// Status of saving.
typedef uint32 RecordState;

/// Start a new record.
static void
startRecord(Stream& s, int type, RecordState& state)
{
    state = s.getPos();

    char header[6];
    storeInt16(header, type);
    storeInt32(header+2, 0);
    if (s.write(header, sizeof(header)) != sizeof(header))
        throw FileIOException(s, _("Write error"));
}

/// End a record.
static void
endRecord(Stream& s, RecordState& state)
{
    // state points to beginning of header. Since then, we wrote 6+N bytes.
    uint32 pos = s.getPos();
    uint32 bytes_in_chunk = s.getPos() - state - 6;

    // write it
    char tmp[4];
    storeInt32(tmp, bytes_in_chunk);
    s.seek(state + 2);
    if (s.write(tmp, sizeof(tmp)) != sizeof(tmp))
        throw FileIOException(s, _("Write error"));

    // go back
    s.seek(pos);
}

/** Save starchart database.
    \param s   Stream to save into
    \param trn Turn to save */
void
saveChartDatabase(Stream& s, GGameTurn& trn)
{
    // FIXME: this is incomplete
    char header[16] = "CCchart\032";
    storeInt16(header+8, trn.getTurnNumber());
    storeInt16(header+10, 16 /* start offset */);
    if (s.write(header, sizeof(header)) != sizeof(header))
        throw FileIOException(s, _("Write error"));

    const GUniverse& univ = trn.getCurrentUniverse();
    RecordState rs;

    /* Write the drawings. For drawings that have atoms attached, we
       must store the atom mapping first, to be able to restore it. 
       Whereas PCC 1.x stores the actual atom values verbatim, we
       emulate that to support our larger value range. This logic is
       entirely hidden in computeAtomMap(). */
    GDrawingAtomMap map;
    const GDrawingContainer& drawings = univ.getDrawings();
    drawings.computeAtomMap(map, GDrawingContainer::IncludeInvisible);

    if (!map.isEmpty()) {
        startRecord(s, rPaintingTags, rs);
        map.save(s);
        endRecord(s, rs);
    }
    startRecord(s, rPainting, rs);
    drawings.save(s, map);
    endRecord(s, rs);

    /* Write the minefields. It seems PCC 1.x stores only one
       minefield per record. This record has been specified as being
       capable of holding many since ever, and all other CHART.CC
       programs seem to handle that just fine, so we'll happily store
       them all in one. */
    startRecord(s, rMinefield, rs);
    for (GUniverse::MinefieldMap::const_iterator i = univ.getMinefieldBegin(); i != univ.getMinefieldEnd(); ++i) {
        const GMinefield& mf = *i->second;
        TDbMinefield dbm;
        dbm.id    = mf.getId();
        dbm.x     = mf.getPos().x;
        dbm.y     = mf.getPos().y;
        dbm.owner = mf.getOwner();
        dbm.units = mf.getUnitsLastSeen();
        dbm.type  = mf.isWeb();
        dbm.turn  = mf.getTurnLastSeen();
        if (!storeStructure(s, dbm))
            throw FileIOException(s, _("Write error"));
    }
    endRecord(s, rs);

    /* Write autobuild settings */
    startRecord(s, rAutoBuild, rs);
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        TAutobuildSettings abs;
        trn.getPlanet(i).getAutobuild(abs);
        if (!storeStructure(s, abs))
            throw FileIOException(s, _("Write error"));
    }
    endRecord(s, rs);

    /* Write planets */
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        const GPlanet& p = trn.getPlanet(i);
        if (p.hasAnyPlanetData()) {
            TDbPlanet dbp;
            trn.getPlanet(i).getHistoryData(dbp);
            startRecord(s, rPlanetHistory, rs);
            if (!storeStructure(s, dbp))
                throw FileIOException(s, _("Write error"));
            endRecord(s, rs);
        }
    }
}
