/**
  *  \file gfx/anim.h
  *  \brief Animations
  *
  *  Data structures and classes for animated sprites.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_ANIM_H
#define PCC_V2_GFX_ANIM_H

#include <vector>
#include "gfx/pixmap.h"
#include "gfx/sprite3d.h"
#include "cpluslib/assert.h"
#include "cpluslib/smartptr.h"

/** Information about one animation frame. */
struct GfxAniFrame {
    Ptr<GfxPixmap> pixmap;      /**< Pixmap for this frame. */
    int next;                   /**< Index into array, for next frame. negative at end. */
    int time;                   /**< Time this frame is displayed, in ticks. */

    /** Constructor.
        \param pixmap  pixmap to display for this frame
        \param next    number of next frame to display
        \param time    how long to display this frame, in animation ticks */
    GfxAniFrame(Ptr<GfxPixmap> pixmap, int next, int time)
        : pixmap(pixmap), next(next), time(time) { }
};

class GfxAnimation {
    std::vector<GfxAniFrame> data;  ///< All frames.
    int wi, he;                     ///< Size of animation.
 public:
    GfxAnimation();
    ~GfxAnimation();
    void addFrame(Ptr<GfxPixmap> apix, int anext, int atime);

    /** Get frame i. */
    GfxAniFrame& getFrame(int i)
        {
            ASSERT (i >= 0 && unsigned(i) < data.size());
            return data[i];
        }
    /** Get index of first frame. The first frame to display is
        getFrame(getStart()). Currently always returns zero. This is
        for people using the animation, not for those building it. */
    int getStart() const
        { return 0; }
    /** Get width. This is derived from the animation pixmaps. */
    int getWidth() const
        { return wi; }
    /** Get height. This is derived from the animation pixmaps. */
    int getHeight() const
        { return he; }
    /** Get number of frames in this container. */
    int getNumFrames() const
        { return data.size(); }
};

class GfxAniSprite : public GfxSprite3D {
    Ptr<GfxAnimation> anim;     /**< Animation to display. */
    int ip;                     /**< "instruction pointer". */
    int time;                   /**< Remaining time for current frame. */
    void setIndex(int aip);
 public:
    GfxAniSprite(int x, int y, int z, Ptr<GfxAnimation> aanim);
    void tick();
    void draw(GfxCanvas& can);
};

#endif
