#!/usr/bin/perl -w
#
#  Compile message file into binary format
#
#  By using our own message compiler, we can define a portable file format,
#  and can tailor our format string validation to our format() routine.
#

use strict;
use vars qw{$line $lineno $beg_lineno $token $token_arg %strings $input $output $opt_picky $opt_allfmt};

sub usage {
    fail ("usage: $0 [--opts] input output\n   or: $0 languagecode\n\n".
          "Options:\n".
          "  --picky       display more warnings\n".
          "  --all-format  validate all strings as format strings\n");
}

$opt_picky = 0;
$opt_allfmt = 0;

foreach (@ARGV) {
    if (/^--picky$/) {
        $opt_picky = 1
    } elsif (/^--all-format$/) {
        $opt_allfmt = 1;
    } elsif (/^-/) {
        usage();
    } elsif (!defined $input) {
        $input = $_;
    } elsif (!defined $output) {
        $output = $_;
    } else {
        fail ("usage: $0 input output");
    }
}
if (defined $input && !defined $output && $input !~ /[^a-z0-9_]/i) {
    $output = "resource/$input.lang";
    $input = "po/$input.po";
}
if (!defined $output) {
    usage();
}

open INPUT, "< $input" or fail ("can't open $input: $!");

################################ Read Input ###############################

$line = "";
$lineno = 0;
$beg_lineno = 0;
%strings = ();

lex ();
while (defined $token) {
    my $args = '';
    if ($token eq '#,') {
        $args = $token_arg;
        lex ();
    }
    if ($token eq '#~') {
        lex();
        next;
    }
    expect ('msgid');
    $beg_lineno = $lineno;
    my $key = expect_str ();
    expect ('msgstr');
    my $value = expect_str ();
    warning ("Duplicate message id") if exists $strings{$key};
    if ($opt_picky && $args !~ /c-format/ && $key =~ /%[0-9]*[sd]/) {
        warning ("Looks like format string but not marked as one");
    }
    next if $value eq "" || $value eq $key;
    if ($args =~ /c-format/ || $opt_allfmt) {
        validate_format ($key, $value);
    }
    if ($args =~ /fuzzy/) {
        warning ("Fuzzy translation");
    }
    $strings{$key} = $value;
}

############################# Generate Output #############################

my $in     = "";
my $out    = "";
my $inptr  = "";
my $outptr = "";
foreach (sort keys %strings) {
    my ($instr, $outstr) = ("$_\0", "$strings{$_}\0");
    $inptr .= pack 'VV', length ($in), length ($instr);
    $outptr .= pack 'VV', length ($out), length ($outstr);
    $in .= $instr;
    $out .= $outstr;
}
open OUT, "> $output" or &fail ("can't create $output: $!");
binmode OUT;
print OUT "CClang0\032",
    pack ("VVVVVVV",
          length ($inptr) / 8,  # number of entries
          36,                   # input ptr pos
          36 + length ($inptr), # output ptr pos
          36 + length ($inptr) + length ($outptr),    # input data pos
          length ($in),         # input data length
          36 + length ($inptr) + length ($outptr) + length ($in),
          length ($out));
print OUT $inptr, $outptr, $in, $out;
close OUT;
exit 0;

############################### Subroutines ###############################

sub validate_format {
    my ($orig, $new) = @_;
    my $orig_phs = parse_format ('original', $orig);
    my $new_phs = parse_format ('translation', $new);

    my $count = @$orig_phs;
    if (@$orig_phs < @$new_phs) {
        warning ("Translation uses more parameters than original");
    } elsif (@$orig_phs > @$new_phs) {
        warning ("Translation does not use all parameters of original");
        $count = @$new_phs;
    }

    for (my $i = 0; $i < $count; ++$i) {
        my $o = defined $orig_phs->[$i] ? join(',', sort keys %{$orig_phs->[$i]}) : 'none';
        my $t = defined $new_phs->[$i]  ? join(',', sort keys %{$new_phs->[$i]})  : 'none';
        if ($o ne $t) {
            warning ("Translation uses different format specifiers: '$o' -> '$t'");
        }
    }
}

sub parse_format {
    my ($prefix, $str) = @_;
    my $iflevel = 0;
    my $conditions_had_else = 0;
    my @condition_argptr_exit;
    my @condition_argptr_start;
    my $arg_index = 0;
    my @arg_types;
    my $last_test;
    while ($str =~ m{%([0-9\$\- \+\'!\.]*)(.)}g) {
        my ($p, $c) = ($1, $2);
        if ($c eq '{') {
            if ($p !~ m|^!?\d+$|) { warning ("$prefix: strange condition"); }
            if (++$iflevel > 32) { warning ("$prefix: too many nested ifs"); }
            if (!defined $last_test) {
                warning ("$prefix: '%{' without preceding output");
            } else {
                $arg_types[$last_test]{cond} = 1
                  if $opt_picky;
            }
            $conditions_had_else <<= 1;
            push @condition_argptr_start, $arg_index;
            push @condition_argptr_exit, undef;
        } elsif ($c eq '}') {
            if (!$iflevel) { warning ("$prefix: '%}' without '%{'") } else { --$iflevel; }
            if ($p ne '') { warning ("$prefix: '%}' does not take arguments"); }
            if ($conditions_had_else & 1) {
                # We had an 'else', so validate that positions match
                if ($condition_argptr_exit[-1] != $arg_index) { $arg_index = undef }
            }
            $conditions_had_else >>= 1;
            pop @condition_argptr_exit;
            pop @condition_argptr_start;
        } elsif ($c eq '|') {
            if ($iflevel == 0) { warning ("$prefix: '%|' without '%{'") }
            if ($p ne '') { warning ("$prefix: '%|' does not take arguments"); }
            if ($conditions_had_else & 1) {
                warning ("$prefix: duplicate 'else'");
            } else {
                $conditions_had_else |= 1;
                $condition_argptr_exit[-1] = $arg_index;
                $arg_index = $condition_argptr_start[-1];
            }
        } elsif ($c eq '%') {
            if ($p ne "") { warning ("$prefix: '%%' does not take arguments"); }
        } else {
            if ($p =~ s{^([0-9]+)\$}{}) {
                $arg_index = $1;
            }
            if (!defined $arg_index) {
                warning ("$prefix: argument pointer not known for '%$p$c'");
            } else {
                $last_test = $arg_index++;
                while (@arg_types < $arg_index) {
                    push @arg_types, {};
                }
                $arg_types[$last_test]{$c} = 1;
            }
        }
    }
    if ($iflevel != 0) {
        warning ("$prefix: $iflevel conditionals not closed at end");
    }
    \@arg_types;
}

sub expect_str {
    my $rv;
    while (defined $token && $token eq '"') {
        $rv .= $token_arg;
        lex ();
    }
    $rv;
}

sub expect {
    my $rv;
    foreach (@_) {
        $rv = $token_arg;
        if ($token ne $_) {
            fail ("expected `$_'");
        }
        lex ();
    }
    $rv;
}

sub lex {
    while (1) {
        if ($line eq "") {
            $line = <INPUT>;
            return ($token = undef) if !defined $line;
            ++$lineno;
            chomp $line;
            next;
        }
        if ($line =~ s/^\s+//) {
            next;
        } elsif ($line =~ s/^\#,\s*(.*)//) {
            ($token, $token_arg) = ("#,", $1);
            return $token;
        } elsif ($line =~ s/^\#~\s*(.*)//) {
            ($token, $token_arg) = ("#~", $1);
            return $token;
        } elsif ($line =~ s/^\#.*//) {
            next;
        } elsif ($line =~ s/^(\w+)//) {
            ($token, $token_arg) = ($1, "");
            return $token;
        } elsif ($line =~ s/^\"(([^\"\\]|\\.)*)\"//) {
            my $text = $1;
            $text =~ s/\\([0-7]{1,3}|.)/unescape ($1)/eg;
            ($token, $token_arg) = ('"', $text);
            return $token;
        } else {
            fail ("syntax error (unable to parse)");
        }
    }
}

sub unescape {
    my $what = shift;
    $what =~ /^[0-7]*$/ and return chr(oct($what));
    $what eq "n" and return "\n";
    $what eq "a" and return "\7";
    $what eq "\"" and return "\"";
    $what eq "\'" and return "\'";
    $what eq "\\" and return "\\";
    warning ("unknown escape sequence `\\$what'");
    return "?";
}

sub warning {
    print STDERR "$input:$beg_lineno: warning: @_\n";
}

sub fail {
    if (defined $lineno) {
        print STDERR "$input:$lineno: @_\n";
    } else {
        print STDERR "$0:@_\n";
    }
    exit 1;
}
