/**
  *  \file ui/layout-flow.cc
  *  \brief Implementation of the Flow Layout
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/layout.h"

/** \class UIFlowLayout
    \brief Flow Layout

    This layout manager positions widgets in a row, with line breaks
    if needed, much like a typesetter positions words. It can allocate
    from top-left to bottom-right, or vice versa.

    Unlike the Java FlowLayout, you have to specify the number of
    lines you expect to use (so getLayoutInfo() can report the correct
    size). Like the Java FlowLayout, however, this might exceed the
    allocated size anyway, though. */

void
UIFlowLayout::doLayout(UIWidgetContainer& container, GfxRect size)
{
    UIWidget::LayoutInfo info;
    getLayoutInfo(container, info);

    GfxPoint UIWidget::LayoutInfo::*which_size = (size.w < info.pref_size.x
                                                  ? &UIWidget::LayoutInfo::min_size
                                                  : &UIWidget::LayoutInfo::pref_size);
    /* reverse of the computation in getLayoutInfo */
    int line_height = ((info.*which_size).y - (vgap * (num_lines-1))) / num_lines;
    int xoffs = 0, yoffs = 0;

    for (UIWidget* p = (right_just ? container.getLastWidget() : container.getFirstWidget());
         p != 0;
         p = (right_just ? p->getPrev() : p->getNext()))
    {
        p->getLayoutInfo(info);
        if (info.flags & UIWidget::li_NoLayout)
            continue;

        if (xoffs + (info.*which_size).x > size.w) {
            /* new line */
            yoffs += line_height + vgap;
            xoffs = 0;
        }

        if (right_just)
            p->setExtent(GfxRect(size.x + size.w - xoffs - (info.*which_size).x,
                                 size.y + size.h - yoffs - (info.*which_size).y,
                                 (info.*which_size).x,
                                 (info.*which_size).y),
                         true);
        else
            p->setExtent(GfxRect(size.x + xoffs, size.y + yoffs,
                                 (info.*which_size).x, (info.*which_size).y),
                         true);
        p->doLayout();

        xoffs += (info.*which_size).x + hgap;
    }
}

void
UIFlowLayout::getLayoutInfo(UIWidgetContainer& container, UIWidget::LayoutInfo& info)
{
    /* compute the maximum of minimum/preferred height, and simply return that. */
    info.flags = UIWidget::li_FlexibleH;
    info.min_size = info.pref_size = GfxPoint(0, 0);
    int num_widgets = 0;
    int total_width = 0, total_pref_width = 0;
    for (UIWidget* w = container.getFirstWidget(); w != 0; w = w->getNext()) {
        UIWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        if (winfo.flags & UIWidget::li_NoLayout)
            continue;
        ++num_widgets;
        if (winfo.min_size.x > info.min_size.x)
            info.min_size.x = winfo.min_size.x;
        if (winfo.min_size.y > info.min_size.y)
            info.min_size.y = winfo.min_size.y;
        if (winfo.pref_size.x > info.pref_size.x)
            info.pref_size.x = winfo.pref_size.x;
        if (winfo.pref_size.y > info.pref_size.y)
            info.pref_size.y = winfo.pref_size.y;
        total_width += winfo.min_size.x;
        total_pref_width += winfo.pref_size.x;
    }

    /* now, we have in info.min_size the minimum size of the largest widget,
       likewise, in info.pref_size the preferred size of the largest one.

       Our height simply is N-times that.

       For the width: assume the worst. We have /total_width/ pixels
       to allocate in /num_lines/ lines, yielding a width of
       /total_width / num_lines/. We must however assume that the largest
       widgets always look over the end. Like this:
            xx xxxx xxxxxxxxxxxx
            xxxx x xxxxxxxxxxxx
            x xx xx xxxxxxxxxxxx
            \        \
               total_width/num_lines
       In addition, total_width actually is total_width + (num_widgets-num_lines)*hgap

       This is a pessimistic assumption. On the plus side, a dialog will look
       much better if there's some space remaining besides the button bar :)
    */
    info.min_size.y  = info.min_size.y  * num_lines + vgap * (num_lines-1);
    info.pref_size.y = info.pref_size.y * num_lines + vgap * (num_lines-1);
    if (num_widgets > num_lines) {
        total_width += (num_widgets - num_lines) * hgap;
        total_pref_width += (num_widgets - num_lines) * hgap;
        if (num_lines > 1) {
            info.min_size.x = total_width / num_lines + info.min_size.x;
            info.pref_size.x = total_pref_width / num_lines + info.pref_size.x;
        } else {
            info.min_size.x = total_width;
            info.pref_size.x = total_pref_width;
        }
    }
}
