/**
  *  \file util/conf.h
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_CONF_H
#define PCC_V2_UTIL_CONF_H

#include <vector>
#include "cpluslib/string.h"
#include "cpluslib/assert.h"
#include "cpluslib/types.h"

class ConfigOption;

/** Configuration object. An object of this type contains parser and
    value information for a configuration.

    \todo How do we make temporary copies of such an object and/or
    transfer data between two of them? */
class Config {
    std::vector<ConfigOption*> options;

    Config(const Config&);
    Config& operator=(const Config&);
 public:
    Config();
    ~Config();
    void addConfigOption(ConfigOption& option);
    void markAllOptionsUnset();
    ConfigOption* getConfigOption(string_t name);
};

/** Configuration option, base class. This is the basic interface to
    assigning configuration options. */
class ConfigOption {
    const char* name;
    bool was_set;
 public:
    ConfigOption(Config& container, const char* name);
    virtual ~ConfigOption()
        { }
    virtual void set(string_t value) = 0;
    const char* getName() const;
    bool wasSet() const
        { return was_set; }
    void markSet()
        { was_set = true; }
    void markUnset()
        { was_set = false; }
};

/** Value parser, base class. Descendants of this class parse strings
    into integers according to particular rules. */
class ValueParser {
 public:
    virtual ~ValueParser()
        { }
    virtual int32 parse(string_t value) = 0;
    void parseArray(string_t value, int32 array[], int dim);
};

/** Value parser for integers. This parses regular decimal integers. */
class ValueIntParser : public ValueParser {
 public:
    ValueIntParser();
    int32 parse(string_t value);
};

/** Integer option. This contains a single int32 value, parsed from
    text according to a ValueParser. */
class ConfigIntOption : public ConfigOption {
    int32 value;
    ValueParser& parser;
 public:
    ConfigIntOption(Config& container, const char* name, ValueParser& parser);
    void set(string_t value);
    void set(int32 new_value)
        { value = new_value; }
    int32 operator()() const
        { return value; }
    void copyFrom(const ConfigIntOption& other)
        { value = other.value; }
};

/** Integer option array. This contains an array of int32 values,
    parsed from a comma-separated list according to a ValueParser. */
template<int N>
class ConfigIntArrayOption : public ConfigOption {
    int32 values[N];
    ValueParser& parser;
 public:
    ConfigIntArrayOption(Config& container, const char* name, ValueParser& parser)
        : ConfigOption(container, name),
          parser(parser)
        { set(0); }
    void set(string_t value);
    void set(int index, int value);
    void set(int value);
    int32 operator()(int index) const;
    void copyFrom(const ConfigIntArrayOption<N>& other);
};

/** String option. This contains a single string. */
class ConfigStringOption : public ConfigOption {
    string_t value;
 public:
    ConfigStringOption(Config& container, const char* name);
    ~ConfigStringOption();
    void set(string_t value);
    string_t operator()() const;
};

/** Alias option. This simply forwards all requests to another option. */
class ConfigAlias : public ConfigOption {
    ConfigOption& other;
 public:
    ConfigAlias(Config& container, const char* name, ConfigOption& other);
    void set(string_t value);
};

/************************** ConfigIntArrayOption *************************/

template<int N>
void
ConfigIntArrayOption<N>::set(string_t value)
{
    parser.parseArray(value, this->values, N);
}

template<int N>
void
ConfigIntArrayOption<N>::set(int index, int value)
{
    ASSERT(index > 0 && index <= N);
    this->values[index-1] = value;
}

template<int N>
void
ConfigIntArrayOption<N>::set(int value)
{
    for (int i = 0; i < N; ++i)
        this->values[i] = value;
}

template<int N>
int32
ConfigIntArrayOption<N>::operator()(int index) const
{
    // if index is out of bounds, return last value (i.e. Colony). This makes
    // more sense then returning the first (Fed), which differs in more ways
    // from standard than colony, at least for options important to us.
    if(index > 0 && index <= N)
        return this->values[index-1];
    else
        return this->values[N-1];
}

template<int N>
void
ConfigIntArrayOption<N>::copyFrom(const ConfigIntArrayOption<N>& other)
{
    for (int i = 0; i < N; ++i)
        this->values[i] = other.values[i];
}

#endif
