/**
  *  \file util/filepars.cc
  *  \brief File Parser
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include <cctype>
#include "util/filepars.h"
#include "cpluslib/string.h"
#include "util/misc.h"

/** \class FileParser
    \brief Abstract File Parser Class

    This class provides the logic to read text files. It will split
    the file into lines, wield out empty and comment lines, and call
    process() or unparsed() for each line. Derived classes will
    override these methods to actually do some parsing.

    Comments are started by using a ';' as first non-blank character.

    In addition, this provides methods to read files from game or root
    directory. */

/** Constructor. Does nothing special. */
FileParser::FileParser()
    : name("<unnamed>"), comment_char(';')
{ }

/** Callback: parse a text line. This function is called for each line
    which is neither empty nor a comment.
    \param line_id     line number
    \param line        line text */
void
FileParser::process(int /*line_id*/, string_t /*line*/)
{ }

/** Callback: parse commetn line. This function is called for each
    line which is either empty or a comment. */
void
FileParser::unparsed(int /*line_id*/, string_t /*line*/)
{ }

/** Parse a file. If you wish to use getFileName(), you must set the file
    name before calling this.
    \param in       stream to read from
    \param line_id  the line_id to pass to parser functions. The first line
                    will be called with line_id+1, the second with line_id+2,
                    and so on. */
void
FileParser::parseFile(std::istream& in, int line_id)
{
    using namespace std;
    string_t line;
    string_t::size_type n;
    while (getline(in, line)) {
        ++line_id;
        if (!line.empty() && line[line.length()-1] == '\r')
            line.erase(line.length()-1);
        for (n = 0; n < line.length() && isspace(n); ++n)
            ;
        if (n >= line.length() || line[n] == comment_char)
            unparsed(line_id, line);
        else
            process(line_id, line);
    }
}

/** Trim comments. If the line contains a comment, remove it. */
void
FileParser::trimComments(string_t& line)
{
    using namespace std;
    string_t::size_type n = line.find(comment_char);
    if (n != string_t::npos) {
        while (n > 0 && isspace(line[n-1]))
            --n;
        line.erase(n);
    }
}
